# Copyright (C) 2010 by Aaron Bentley
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


"""The Kraken: a tool for releasing software on Launchpad."""


doc = """\
INTRODUCTION
The Kraken helps you release software by automating steps.  You run it in a
Bazaar branch, and it automatically determines the related project and series.
If you confirm, it creates a tarball of the current branch and uploads it to
Launchpad, and then tags the branch appropriately.

If you create a file named .the_kraken/changelog, it will be used for the
release changelog field.  The first line of this file must be the intended
version, and the rest should be normal text.

CONFIGURATION
The Kraken requires a small config file named ".the_kraken/config" in the
working directory.  It is formatted using ini style.

CONFIGURATION VARIABLES
testing (boolean, mandatory)
Controls whether The Kraken runs in testing mode and uploads to staging.

require_code_name (boolean)
If true, require a code name to be specified with --code-name.

series (string)
The name of the Launchpad series to use.  Normally, this is autodetected from
the branch, but can be specified if multiple series link to a branch.

edit_changelog (boolean)
If true, let the user edit the changelog during --prepare.  The Kraken
prefers the 'editor' configuration variable, but falls back to the EDITOR
environment variable and 'vi' if EDITOR is not set.

editor (string)
Name and initial arguments of the editor to use.  (e.g. 'gvim -f')

check_script
A command to run to perform project-specific checks.  If this script returns
nonzero, processing halts.  (e.g. './check-release.py')

--------------example config-----------------
testing = true
require_code_name = true
---------------------------------------------

Release tarballs and signatures are stored in ~/releases.
"""

__version__ = '1.0.1dev'


from contextlib import contextmanager
import datetime
import errno
import os.path
from subprocess import call
import shlex
import shutil
import sys
import tempfile
import webbrowser

from bzrlib import branch
from bzrlib import export
from bzrlib.plugins.launchpad import lp_api
from bzrlib import osutils
from configobj import ConfigObj
from launchpadlib.launchpad import Launchpad
from launchpadlib.uris import LPNET_SERVICE_ROOT, STAGING_SERVICE_ROOT


RELEASE_DIR = '~/releases'


@contextmanager
def temp_dir():
    old_dir = os.getcwd()
    tempdir = tempfile.mkdtemp()
    os.chdir(tempdir)
    try:
        yield tempdir
    finally:
        os.chdir(old_dir)
        shutil.rmtree(tempdir)


class UserError(Exception):
    pass


class LaunchpadBranch(lp_api.LaunchpadBranch):

    @staticmethod
    def tweak_url(url, launchpad):
        """Adjust a URL to work with staging, if needed."""
        if not str(launchpad._root_uri).startswith(STAGING_SERVICE_ROOT):
            return url
        if url is None:
            return None
        return url.replace('bazaar.launchpad.net',
                           'bazaar.staging.launchpad.net')


class TheKrakenData(object):

    CHANGELOG_PATH = '.the_kraken/changelog'

    @staticmethod
    def get_config():
        CONFIG_PATH = '.the_kraken/config'
        if not os.path.exists(CONFIG_PATH):
            raise UserError('Config file "%s" not found' % CONFIG_PATH)
        return ConfigObj(CONFIG_PATH)

    def set_changelog(self, changelog_version, changelog):
        if '\n' in changelog_version:
            raise ValueError('Newline not allowed in version.')
        with open(self.CHANGELOG_PATH, 'w') as changelog_file:
            changelog_file.write(changelog_version + '\n')
            changelog_file.write(changelog)


    def get_changelog(self):
        try:
            changelog_file = open(self.CHANGELOG_PATH)
        except IOError as e:
            if e.errno == errno.ENOENT:
                return None, None
            else:
                raise
        else:
            with changelog_file:
                changelog_version = changelog_file.readline().rstrip()
                return changelog_version, changelog_file.read()


class TheKraken(object):

    def __init__(self, location, version, config, code_name=None,
                 _launchpad=None):
        self.data = TheKrakenData()
        self.location = location
        self.version = version
        self.code_name = code_name
        self.changelog_version, self.changelog = self.data.get_changelog()
        self.tag = 'release-%s' % version
        self.config = config
        self.require_code_name = self._get_bool('require_codename')
        self.edit_changelog = self._get_bool('edit_changelog')
        self.testing = config.as_bool('testing')
        self.check_script = config.get('check_script')
        self.series_name = config.get('series')
        self._launchpad = _launchpad

    def _get_bool(self, name, default=None):
        if name in self.config:
            return self.config.as_bool(name)
        else:
            if default is None:
                return False
            else:
                return default

    @property
    def launchpad(self):
        if self._launchpad is None:
            if self.testing:
                root = STAGING_SERVICE_ROOT
            else:
                root = LPNET_SERVICE_ROOT
            self._launchpad = Launchpad.login_with('The Kraken', root)
        return self._launchpad

    def get_branch(self):
        bzr_branch = branch.Branch.open(self.location)
        return LaunchpadBranch.from_bzr(self.launchpad, bzr_branch)

    @classmethod
    def get_bzr_changes(cls, bzr_branch):
        changes = []
        revision_ids = cls._get_release_revids(bzr_branch)
        revisions = dict((r.revision_id, r) for r
                         in bzr_branch.repository.get_revisions(revision_ids))
        for revision_id in revision_ids:
            message = revisions[revision_id].message
            if not message.endswith('\n'):
                message += '\n'
            changes.append(message)
        return '\n'.join(changes)

    @staticmethod
    def _get_release_revids(bzr_branch):
        revision_ids = bzr_branch.revision_history()
        rev_tag_dict = bzr_branch.tags.get_reverse_tag_dict()
        repo = bzr_branch.repository
        last_revision = bzr_branch.last_revision()
        revision_ids = repo.iter_reverse_revision_history(last_revision)
        release_ids = []
        for revision_id in revision_ids:
            tags = rev_tag_dict.get(revision_id, [])
            revision_tags = [tag for tag in tags if tag.startswith('release-')]
            if len(revision_tags) > 0:
                break
            release_ids.append(revision_id)
        release_ids.reverse()
        return release_ids

    def get_series(self, lpb):
        if self.series_name is not None:
            for series in lpb.lp.project.series:
                if series.name == self.series_name:
                    return series
            else:
                raise UserError('No series named %s' % self.series_name)

        series = [s for s in lpb.lp.project.series if s.branch == lpb.lp]
        if len(series) == 0:
            raise UserError('No series linked to branch "%s".' %
                            lpb.lp.bzr_identity)
        if len(series) > 1:
            raise UserError('Multiple series linked to branch "%s".' %
                            lpb.lp.bzr_identity)
        return series[0]

    def upload_tarball(self, release, tarball_path):
        with open(tarball_path) as tarball:
            tarball_content = tarball.read()
        with open(tarball_path + '.sig') as sig:
            sig_content = sig.read()
        tarball_name = os.path.basename(tarball_path)
        release.add_file(file_type='Code Release Tarball',
            file_content=tarball_content, filename=tarball_name,
            signature_content=sig_content,
            signature_filename=tarball_name + '.sig',
            content_type="application/x-gzip; charset=binary")

    def get_summary(self, project_name, series, tarball_path):
        yield "  Testing: %s\n" % self.testing
        yield "  Project: %s\n" % project_name
        yield "   Series: %s\n" % series.name
        yield "  Version: %s\n" % self.version
        yield "Code name: %s\n" % self.code_name
        yield "  bzr tag: %s\n" % self.tag
        if self.changelog:
            yield "Changes:\n"
            yield self.changelog

    def get_tarball_path(self, project_name):
        tarball_name = '%s-%s.tar.gz' % (project_name, self.version)
        if self.testing:
            tarball_name = 'testing-' + tarball_name
        release_dir = os.path.expanduser(RELEASE_DIR)
        tarball_path = os.path.join(release_dir, tarball_name)
        return tarball_path

    @staticmethod
    def write_tarball(tarball_path, branch, project_name):
        export.export(branch.bzr.basis_tree(), tarball_path, root=project_name)
        call(['gpg', '--detach-sign', tarball_path])

    def create_release(self, series):
        milestone = series.newMilestone(name=self.version,
                                        code_name=self.code_name)
        return milestone.createProductRelease(
            date_released=datetime.datetime.utcnow(), changelog=self.changelog)

    def prepare(self):
        bzr_branch = branch.Branch.open(self.location)
        bzr_branch.lock_write()
        try:
            changes = self.get_bzr_changes(bzr_branch)
        finally:
            bzr_branch.unlock()
        if self.edit_changelog:
            changes = self._user_edit(changes)
        self.data.set_changelog(self.version, changes)

    def _editor_command(self):
        editor = self.config.get('editor')
        if editor is None:
            editor = os.environ.get('EDITOR')
        if editor is None:
            editor = 'vi'
        return shlex.split(editor)

    def _user_edit(self, text):
        with temp_dir() as text_dir:
            text_path = os.path.join(text_dir, 'file.txt')
            with open(text_path, 'w') as text_file:
                text_file.write(text)
            call(self._editor_command() + [text_path])
            with open(text_path, 'r') as text_file:
                return text_file.read()

    def run(self):
        lpb = self.get_branch()
        series = self.get_series(lpb)
        project_name = lpb.lp.project.name
        tarball_path = self.get_tarball_path(project_name)
        summary = self.get_summary(project_name, series, tarball_path)
        sys.stdout.writelines(summary)
        self.check(tarball_path, lpb)
        print "Continue? [y/n]"
        ch = osutils.getchar()
        if ch.lower() != 'y':
            sys.exit(0)
        release = self.finish_release(lpb, project_name, series, tarball_path)
        webbrowser.open(self.get_release_url(release))

    def check(self, tarball_path, lpb):
        if self.changelog_version is not None:
            if self.changelog_version != self.version:
                raise UserError('Changelog is for version %s' %
                                self.changelog_version)
        if self.require_code_name and self.code_name is None:
            raise UserError('Code name not specified.')
        if os.path.exists(tarball_path):
            raise UserError('File exists: "%s"' % tarball_path)
        if lpb.bzr.tags.has_tag(self.tag):
            raise UserError('Tag already exists: %s' % self.tag)
        if self.check_script is not None:
            status = call(shlex.split(self.check_script))
            if status != 0:
                message = '"%s" exited with status %d' % (self.check_script,
                                                          status)
                raise UserError(message)

    def finish_release(self, lpb, project_name, series, tarball_path):
        self.write_tarball(tarball_path, lpb, project_name)
        release = self.create_release(series)
        self.upload_tarball(release, tarball_path)
        tag_name = 'release-%s' % self.version
        lpb.bzr.tags.set_tag(self.tag, lpb.bzr.last_revision())
        hosted_branch = branch.Branch.open(lpb.lp.bzr_identity)
        lpb.bzr.push(hosted_branch)

        return release

    @staticmethod
    def get_release_url(release):
        release_url = release.self_link
        release_url = release_url.replace('https://api.', 'https://')
        return release_url.replace('launchpad.net/1.0', 'launchpad.net')
