# Copyright (C) 2010 by Aaron Bentley
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


from contextlib import contextmanager
import os
import re
from textwrap import dedent
import unittest

from bzrlib import (
    bzrdir,
    transform,
)
from bzrlib.branch import Branch
from bzrlib.transport import memory

from configobj import ConfigObj

from the_kraken import (
    LaunchpadBranch,
    RELEASE_DIR,
    temp_dir,
    TheKraken,
    TheKrakenData,
    UserError,
)


class FakeProject(object):

    def __init__(self):
        self.series = []


class FakeMilestone(object):

    def __init__(self, name, code_name=None):
        self.name = name
        self.code_name = code_name

    def createProductRelease(self, date_released, changelog=None):
        return FakeRelease(date_released, milestone=self, changelog=changelog)


class FakeSeries(object):

    def __init__(self, branch, name=None):
        self.branch = branch
        self.name = name

    def newMilestone(self, name, code_name):
        return FakeMilestone(name, code_name)


class FakeRelease(object):

    def __init__(self, date_released, self_link=None, milestone=None,
                 changelog=None):
        self.files = []
        self.self_link = self_link
        self.milestone = milestone
        self.changelog = changelog

    def add_file(self, file_type, file_content, filename, signature_content,
            signature_filename, content_type):
        self.files.append({
            'file_type': file_type,
            'file_content': file_content,
            'filename': filename,
            'signature_content': signature_content,
            'signature_filename': signature_filename,
            'content_type': content_type,
        })


class FakeBranch(object):

    def __init__(self, bzr_identity, project):
        self.bzr_identity = bzr_identity
        self.project = project

class FakeBranches(object):

    def __init__(self, branch_urls=None):
        if branch_urls is None:
            self.urls = {}
        else:
            self.urls = branch_urls

    def getByUrl(self, url):
        return self.urls.get(url)


class FakeLaunchpad(object):

    _root_uri = 'https://api.staging.launchpad.net/1.0/'

    def __init__(self, branch_urls=None):
        self.branches = FakeBranches(branch_urls)


class TestLaunchpadBranch(unittest.TestCase):
    """Test local LaunchpadBranch."""

    def test_tweak_url(self):
        tweaked = LaunchpadBranch.tweak_url('bzr+ssh://bazaar.launchpad.net',
                                            FakeLaunchpad)
        self.assertEqual('bzr+ssh://bazaar.staging.launchpad.net', tweaked)


class MemoryServer(memory.MemoryServer):

    def __enter__(self):
        self.start_server()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop_server()


@contextmanager
def expected_exception(cls, str_re=None):
    try:
        yield
    except cls as e:
        if str_re is not None:
            if not re.match(str_re, str(e)):
                raise AssertionError('"%s" does not match "%s".' %
                                     (str(e), str_re))
    else:
        raise AssertionError('"%s" was not raised.' % cls.__name__)


class TestTheKrakenData(unittest.TestCase):
    """Test TheKrakenData"""

    def test_get_config(self):
        with temp_dir():
            with expected_exception(UserError,
                                    'Config file .* not found'):
                TheKrakenData.get_config()
            os.mkdir('.the_kraken')
            with file('.the_kraken/config', 'w') as config_file:
                config_file.write('testing=False')
            config = TheKrakenData.get_config()
        self.assertEqual(config.as_bool('testing'), False)

    def test_get_changelog(self):
        with temp_dir():
            os.mkdir('.the_kraken')
            with open('.the_kraken/changelog', 'w') as changelog_file:
                changelog_file.write('0.99\nchanges\n')
            data = TheKrakenData()
            changelog_version, changelog = data.get_changelog()
        self.assertEqual('0.99', changelog_version)
        self.assertEqual('changes\n', changelog)

    def test_set_changelog(self):
        with temp_dir():
            os.mkdir('.the_kraken')
            data = TheKrakenData()
            data.set_changelog('0.99', 'changes\n')
            with open('.the_kraken/changelog') as changelog_file:
                self.assertEqual('0.99\nchanges\n', changelog_file.read())

    def test_set_changelog_newline(self):
        with temp_dir():
            os.mkdir('.the_kraken')
            data = TheKrakenData()
            with expected_exception(ValueError,
                                    'Newline not allowed in version.'):
                data.set_changelog('0.99\n', 'changes\n')


class TestTheKraken(unittest.TestCase):
    """Test TheKraken."""

    def make_the_kraken(self, location='.', config=None):
        if config is None:
            config = ConfigObj()
        config['testing'] = 'true'
        branch_url = 'bzr+ssh://bazaar.staging.launchpad.net/foo'
        project = FakeProject()
        launchpad = FakeLaunchpad({
            branch_url: FakeBranch('hosted_branch', project)})
        return TheKraken(location, '1.0', config, 'Virgin version',
                         _launchpad=launchpad)

    def make_branch(self, url):
        branch = bzrdir.BzrDir.create_branch_convenience(url)
        branch.set_push_location('bzr+ssh://bazaar.launchpad.net/foo')
        return branch

    def test_get_branch(self):
        with MemoryServer() as server:
            bzr_branch = self.make_branch(server.get_url())
            kraken = self.make_the_kraken(bzr_branch.base)
            lp_branch = kraken.get_branch().lp
            self.assertEqual('hosted_branch', lp_branch.bzr_identity)

    def test_get_series(self):
        with MemoryServer() as server:
            bzr_branch = self.make_branch(server.get_url())
            kraken = self.make_the_kraken(bzr_branch.base)
            branch = kraken.get_branch()
            with expected_exception(UserError, "No series linked to branch"):
                kraken.get_series(branch)
            series = FakeSeries(branch.lp)
            branch.lp.project.series.append(series)
            self.assertEqual(series, kraken.get_series(branch))
            branch.lp.project.series.append(series)
            series2 = FakeSeries(branch.lp, 'second')
            branch.lp.project.series.append(series2)
            with expected_exception(UserError,
                                    "Multiple series linked to branch"):
                kraken.get_series(branch)
            kraken.series_name = 'third'
            with expected_exception(UserError, "No series named third"):
                kraken.get_series(branch)
            kraken.series_name = 'second'
            self.assertEqual(series2, kraken.get_series(branch))

    def test_get_tarball_path(self):
        with temp_dir():
            kraken = self.make_the_kraken('.')
            release_dir = os.path.expanduser(RELEASE_DIR)
            expected = os.path.join(release_dir, 'testing-projname-1.0.tar.gz')
            self.assertEqual(expected, kraken.get_tarball_path('projname'))
            kraken.testing = False
            expected = os.path.join(release_dir, 'projname-1.0.tar.gz')
            self.assertEqual(expected, kraken.get_tarball_path('projname'))

    def test_get_bzr_changes(self):
        with MemoryServer() as server:
            bzr_branch = self.make_branch(server.get_url())
            bzr_branch.lock_write()
            try:
                last = self.commit(bzr_branch, 'last release')
                bzr_branch.tags.set_tag('release-93', last)
                self.commit(bzr_branch, 'added foo')
                self.commit(bzr_branch, 'changed bar\n')
                self.commit(bzr_branch, 'deleted baz')
                bzr_changes = TheKraken.get_bzr_changes(bzr_branch)
            finally:
                bzr_branch.unlock()
            self.assertEqual(dedent("""\
                added foo

                changed bar

                deleted baz
                """), bzr_changes)


    def test_get_bzr_change_revids(self):
        with MemoryServer() as server:
            bzr_branch = self.make_branch(server.get_url())
            bzr_branch.lock_write()
            try:
                rev1 = self.commit(bzr_branch, 'added foo')
                rev2 = self.commit(bzr_branch, 'changed bar\n')
                rev3 = self.commit(bzr_branch, 'deleted baz')
                bzr_branch.tags.set_tag('release-93', rev1)
                bzr_branch.tags.set_tag('boopa', rev2)
                self.assertEqual([rev2, rev3],
                                 TheKraken._get_release_revids(bzr_branch))
            finally:
                bzr_branch.unlock()

    def test_prepare(self):
        with temp_dir():
            os.mkdir('.the_kraken')
            bzr_branch = self.make_branch('.')
            bzr_branch.lock_write()
            try:
                rev1 = self.commit(bzr_branch, 'added foo')
                rev2 = self.commit(bzr_branch, 'changed bar\n')
            finally:
                bzr_branch.unlock()
            kraken = self.make_the_kraken(bzr_branch.base)
            kraken.prepare()
            self.assertEqual('added foo\n\nchanged bar\n',
                             kraken.data.get_changelog()[1])


    def test_summary(self):
        with temp_dir():
            os.mkdir('.the_kraken')
            TheKrakenData().set_changelog('1.0', 'changes\n')
            kraken = self.make_the_kraken('.')
            summary = kraken.get_summary('projname',
                                         FakeSeries(None, 'fakeseries'),
                                         'tarball-path')
            self.assertEqual('  Testing: True\n'
                             '  Project: projname\n'
                             '   Series: fakeseries\n'
                             '  Version: 1.0\n'
                             'Code name: Virgin version\n'
                             '  bzr tag: release-1.0\n'
                             'Changes:\n'
                             'changes\n', ''.join(summary))

    def test_check_tarball(self):
        with temp_dir():
            self.make_branch('.')
            kraken = self.make_the_kraken('.')
            with expected_exception(UserError, 'File exists: "."'):
                kraken.check('.', kraken.get_branch())

    def test_check_tag(self):
        with temp_dir():
            branch = self.make_branch('.')
            branch.tags.set_tag('release-1.0', 'booga')
            kraken = self.make_the_kraken(branch.base)
            lpb = kraken.get_branch()
            with expected_exception(UserError,
                                    'Tag already exists: release-1.0'):
                kraken.check('foo', lpb)

    def test_check_code_name(self):
        with temp_dir():
            self.make_branch('.')
            config = ConfigObj()
            config['require_code_name'] = 'true'
            kraken = self.make_the_kraken('.', config)
            kraken.code_name = None
            lpb = kraken.get_branch()
            with expected_exception(UserError,
                                    'Code name not specified.'):
                kraken.check('foo', lpb)

    def test_check_changes(self):
        with temp_dir():
            self.make_branch('.')
            os.mkdir('.the_kraken')
            TheKrakenData().set_changelog('0.99', 'changes\n')
            kraken = self.make_the_kraken('.')
            lpb = kraken.get_branch()
            with expected_exception(UserError,
                                    'Changelog is for version 0.99'):
                kraken.check('foo', lpb)
            kraken.version = '0.99'
            kraken.check('foo', lpb)

    def test_check_script(self):
        with temp_dir() as server:
            bzr_branch = self.make_branch('.')
            kraken = self.make_the_kraken(bzr_branch.base)
            kraken.check_script = 'python -c "import sys; sys.exit(1)"'
            lpb = kraken.get_branch()
            with expected_exception(UserError, '.*exited with status 1'):
                kraken.check('1.0.1', lpb)
            kraken.check_script = 'python -c "import sys; sys.exit(0)"'
            kraken.check('1.0.1', lpb)

    def test_write_tarball(self):
        with temp_dir():
            self.make_branch('.')
            kraken = self.make_the_kraken('.')
            tarball_path = 'output.tar.gz'
            kraken.write_tarball(tarball_path, kraken.get_branch(),
                                 'projname')
            self.assertTrue(os.path.exists(tarball_path))
            self.assertTrue(os.path.exists(tarball_path + '.sig'))

    def test_create_release(self):
        with temp_dir():
            kraken = self.make_the_kraken('.')
            release = kraken.create_release(FakeSeries(None))

    def test_create_release_milestone(self):
        with temp_dir():
            kraken = self.make_the_kraken('.')
            release = kraken.create_release(FakeSeries(None))
            self.assertEqual('Virgin version', release.milestone.code_name)

    def test_create_release_changeslog(self):
        with temp_dir():
            os.mkdir('.the_kraken')
            TheKrakenData().set_changelog('1.0', 'Changes')
            kraken = self.make_the_kraken('.')
            release = kraken.create_release(FakeSeries(None))
            self.assertEqual('Changes', release.changelog)

    def test_upload_tarball(self):
        with temp_dir():
            os.mkdir('subdir')
            tarball_path = 'subdir/tarball.tar.gz'
            with open(tarball_path, 'w') as tarball:
                tarball.write('Contents matter little')
            with open(tarball_path + '.sig', 'w') as tarball:
                tarball.write('Signatures matter little')
            release = FakeRelease(None)
            kraken = self.make_the_kraken('.')
            kraken.upload_tarball(release, tarball_path)
        upload = release.files[0]
        self.assertEqual('tarball.tar.gz', upload['filename'])
        self.assertEqual('Contents matter little', upload['file_content'])
        self.assertEqual('Signatures matter little',
                         upload['signature_content'])
        self.assertEqual('tarball.tar.gz.sig', upload['signature_filename'])
        self.assertEqual('application/x-gzip; charset=binary',
                         upload['content_type'])
        self.assertEqual('Code Release Tarball', upload['file_type'])

    def test_get_release_url(self):
        release = FakeRelease(None, 'https://api.launchpad.net/1.0/foo/bar')
        self.assertEqual('https://launchpad.net/foo/bar',
                         TheKraken.get_release_url(release))

    @staticmethod
    def commit(branch, message):
        basis = branch.basis_tree()
        tt = transform.TransformPreview(basis)
        if basis.get_root_id() is None:
            tt.new_directory('', transform.ROOT_PARENT, 'tree-root')
        return tt.commit(branch, message)

    def test_finish_release(self):
        with temp_dir():
            branch = self.make_branch('.')
            branch.lock_write()
            try:
                rev_id = self.commit(branch, 'message')
            finally:
                branch.unlock()
            kraken = self.make_the_kraken('.')
            tarball_path = 'output.tar.gz'
            lpb = kraken.get_branch()
            self.make_branch(lpb.lp.bzr_identity)
            series = FakeSeries(lpb.lp)
            release = kraken.finish_release(lpb, 'projname', series,
                                            tarball_path)
            self.assertTrue(os.path.exists(tarball_path))
            self.assertTrue(os.path.exists(tarball_path + '.sig'))
            self.assertEqual(1, len(release.files))
            tag_items = branch.tags.get_tag_dict().items()
            self.assertEqual([('release-1.0', rev_id)], tag_items)
            hosted_bzr_branch = Branch.open(lpb.lp.bzr_identity)
            self.assertEqual(rev_id, hosted_bzr_branch.last_revision())
            hosted_tag_items = hosted_bzr_branch.tags.get_tag_dict().items()
            self.assertEqual([('release-1.0', rev_id)], hosted_tag_items)
